/*
  CatWeasel MK4 PCI Linux driver (c) 2002-2005 by Dirk Jagdmann <doj@cubic.org>

  This file is part of CatWeasel MK4 PCI Linux driver.

  The CatWeasel MK4 PCI Linux driver is free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  The CatWeasel MK4 PCI Linux driver is distributed in the hope that it
  will be useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
  the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with the CatWeasel MK4 PCI Linux driver; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
  02111-1307 USA
  $Header: /code/cw/device/cwpci.c,v 1.17 2005/04/21 10:27:23 doj Exp $
*/

/*
  Code for finding, initializing and registering the card
*/

#include "cw.h"
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/spinlock.h>

static
#include "firmware.c"

static int cw_num;
int cw_iobase;

/* lock functions ********************************* */

static spinlock_t cw_spinlock=SPIN_LOCK_UNLOCKED;
static unsigned long cw_irq_flags;

void cw_lock(unsigned char a)
{
  spin_lock_irqsave(&cw_spinlock, cw_irq_flags);
  outb(a, cw_iobase + CW_SELECT_BANK);
}
void cw_unlock(void)
{
  spin_unlock_irqrestore(&cw_spinlock, cw_irq_flags);
}

/* ************************************************* */

/* callback function which is called by kernel for every PCI card found */
int __devinit cw_probe(struct pci_dev *pcidev, const struct pci_device_id *pciid)
{
  unsigned int timeout;
  int err;
  unsigned char mem[] = { 1,2,4,8,16,32,64,128,255,44,65,3 };

  if(cw_num>=1)
    {
      printk(KERN_ERR DEVICE_NAME ": found another card, but maximum number (%i) of CatWeasel cards detected.\n", 1);
      return -1;
    }

  if((err=pci_enable_device(pcidev)))
    {
      printk(KERN_ERR DEVICE_NAME ": could not enable device\n");
      return err;
    }

  cw_iobase = pci_resource_start(pcidev, 0);

  if(!request_region(cw_iobase, 256, DEVICE_NAME))
    {
      printk(KERN_ERR DEVICE_NAME ": io ports 0x%04x-0x%04x in use\n", cw_iobase, cw_iobase+255);
      return -EBUSY;
    }

  printk(KERN_NOTICE DEVICE_NAME ": found a CatWeasel at 0x%04X\n", cw_iobase);

  /* Initialize PCI bridge */
  outb(0xf1, cw_iobase + 0x0);
  outb(0x00, cw_iobase + 0x1);
  outb(0xe3, cw_iobase + CW_DATA_DIRECTION);
  outb(CW_BANK_FLOPPY, cw_iobase + CW_SELECT_BANK);
  outb(0x00, cw_iobase + 0x4);
  outb(0x00, cw_iobase + 0x5);
  outb(0x00, cw_iobase + 0x29);
  outb(0x00, cw_iobase + 0x2b);

  /* reset FPGA */
  /*outb(0xe3, cw_iobase + 0x2);*/
  outb(CW_BANK_RESETFPGA, cw_iobase + CW_SELECT_BANK);
  udelay(1000);
  outb(CW_BANK_FLOPPY, cw_iobase + CW_SELECT_BANK);

  /* upload firmware */
  for(err=0; err!=sizeof(cw_firmware); ++err)
    {
      /* set some direction */
      outb( (cw_firmware[err] & 1)?(CW_BANK_FLOPPY+2):CW_BANK_FLOPPY, cw_iobase + CW_SELECT_BANK);
      /* wait for FPGA */
      timeout=0;
      while(!(inb(cw_iobase+7)&8))
	{
	  udelay(1);
	  if(++timeout >= 1000000)
	    {
	      printk(KERN_ERR DEVICE_NAME ": timeout while writing firmware byte %i\n", err);
	      return -EBUSY;
	    }
	}
      /* write byte */
      outb(cw_firmware[err], cw_iobase + CW_FLOPPY_JOYDAT);
    }
#if 0
  /* firmware upload successfull? */
  if(!((err=inb(cw_iobase + 7)) & 4))
    {
      printk(KERN_ERR DEVICE_NAME ": could not reset FPGA: %02x\n", err);
      return -EBUSY;
    }
#endif
  /* now wait until FPGA really comes alive */
  timeout=0;
  while(inb(cw_iobase + CW_FLOPPY_CONTROL) == 10) udelay(1), timeout++;
  printk(KERN_NOTICE DEVICE_NAME ": FPGA needed %ius to start up\n", timeout);

  /* check if memory access is working */
  outb(0, cw_iobase + CW_FLOPPY_RESETPOINTER);
  for(err=0; err!=sizeof(mem); ++err)
    outb(mem[err], cw_iobase + CW_FLOPPY_MEMORY);
  outb(0, cw_iobase + CW_FLOPPY_RESETPOINTER);
  for(err=0; err!=sizeof(mem); ++err)
    if((timeout=inb(cw_iobase + CW_FLOPPY_MEMORY)) != mem[err])
      {
	printk(KERN_ERR DEVICE_NAME ": memory test failed. Byte %i should be %02x but is %02x\n", err, mem[err], timeout);
	return -EBUSY;
      }

  cw_num++;
  return 0;
}

/* PCI cards we are searching for */
static __devinitdata struct pci_device_id id_table[] = {
  {
    .vendor = 0xe159,
    .device = 0x0001,
    .subvendor = 0x5213,
    .subdevice = 0x0003
  },
  { 0 }
};

MODULE_DEVICE_TABLE(pci, id_table);

static struct pci_driver cw_driver = {
  .name = "cw",
  .id_table = id_table,
  .probe = cw_probe,
};

/* module init/exit functions */
void __exit cw_exit(void)
{
  cwjoy_exit();
  amikbd_exit();
  cwsid_exit();

  if(cw_iobase>=0)
    release_region(cw_iobase, 256);
  pci_unregister_driver(&cw_driver);
  printk (KERN_INFO DEVICE_NAME ": uninstalled\n");
}

int __init cw_init(void)
{
  int err;

  cw_num=0;
  cw_iobase=-1;

  err = pci_module_init(&cw_driver);
  if(err)
    {
      cw_exit();
      return err;
    }

  err=cwsid_init();
  if(err<0)
    {
      cw_exit();
      return err;
    }

  err=amikbd_init();
  if(err<0)
    {
      cw_exit();
      return err;
    }

  err=cwjoy_init();
  if(err<0)
    {
      cw_exit();
      return err;
    }

  printk (KERN_INFO DEVICE_NAME ": installed\n");
  return 0;
}

module_init(cw_init);
module_exit(cw_exit);

MODULE_AUTHOR("(c) 2002-2005 Dirk Jagdmann <doj@cubic.org> http://llg.cubic.org/cw");
MODULE_DESCRIPTION("device driver for CatWeasel MK4 PCI");
MODULE_LICENSE("GPL");
