/*
  CatWeasel MK4 PCI Linux driver (c) 2002-2004 by Dirk Jagdmann
  <doj@cubic.org>

  This file is part of CatWeasel MK4 PCI Linux driver.

  The CatWeasel MK4 PCI Linux driver is free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  The CatWeasel MK4 PCI Linux driver is distributed in the hope that
  it will be useful, but WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
  PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with the CatWeasel MK4 PCI Linux driver; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
  MA 02111-1307 USA

  $Header: /code/cw/device/cwsid.c,v 1.28 2005/04/21 10:27:23 doj Exp $
*/

/*
  Code interfacing the SID
*/

#include "cw.h"

#include <asm/io.h>
#include <asm/uaccess.h>
#include <linux/delay.h>

#include <linux/miscdevice.h>
#include <linux/poll.h>

#define CWSID_READBIT	0x20
#define CWSID_NTSCBIT	0x40
#define CWSID_SELECTBIT	0x80

#define CWSID_MAX 2

static int cwsid_ntsc=0;

static int cwsid_delay = 1;
MODULE_PARM(cwsid_delay,"i");
MODULE_PARM_DESC(cwsid_delay,"delay in usec after SID accesses (default=1)");

static unsigned char sid_buffer[CWSID_MAX*0x20];
int sid_typ[CWSID_MAX];

void sid_write(unsigned char addr, unsigned char data, int sid)
{
  if(addr<0x19 && sid<CWSID_MAX)
    {
      outb(data, cw_iobase + CW_FLOPPY_SIDDAT);
      addr|=cwsid_ntsc|((sid&1)?CWSID_SELECTBIT:0);
      outb(addr, cw_iobase + CW_FLOPPY_SIDADDR);
      udelay(cwsid_delay);
    }
}

unsigned char sid_read(unsigned char addr, int sid)
{
  if(addr>=0x19 && addr<=0x1C && sid<CWSID_MAX)
    {
      addr|=cwsid_ntsc|CWSID_READBIT|((sid&1)?CWSID_SELECTBIT:0);
      outb(addr, cw_iobase + CW_FLOPPY_SIDADDR);
      udelay(cwsid_delay);
      return inb(cw_iobase + CW_FLOPPY_SIDDAT);
    }
  return 0;
}

void sid_reset(void)
{
  int i, s;
  cwsid_ntsc=0;
  memset(sid_buffer, 0, sizeof(sid_buffer));

  cw_lock(CW_BANK_FLOPPY);
  for(s=0; s<CWSID_MAX; s++)
    for(i=0; i<0x19; i++)
      sid_write(i, 0, s);
  cw_unlock();
}

void sid_detect(void)
{
  int s;
  unsigned char buf[64], val;

  for(s=0; s!=CWSID_MAX; ++s)
    sid_typ[s]=SID_NONE;

  /* sample pot digitally */
  cw_lock(CW_BANK_IO);
  for(s=0; s!=sizeof(buf); ++s)
    {
      buf[s]=inb(cw_iobase + CW_IO_BUTTON);
      udelay(10);
    }
  cw_unlock();

  /* check for SID0 */
  val=buf[0]&3;
  for(s=1; s!=sizeof(buf); ++s)
    if((buf[s]&3) != val)
      {
	sid_typ[0]=SID_6581;
	break;
      }
  /* check for SID1 */
  val=buf[0]&12;
  for(s=1; s!=sizeof(buf); ++s)
    if((buf[s]&12) != val)
      {
	sid_typ[1]=SID_6581;
	break;
      }

  for(s=0; s!=CWSID_MAX; ++s)
    {
      int i;

      if(sid_typ[s]==SID_NONE) continue;

      /* ripped from hardsid driver */
      sid_reset();
      cw_lock(CW_BANK_FLOPPY);
      sid_write(0x0f, 0xff, s);	/* freq */ //udelay(4);
      sid_write(0x12, 0x30, s);	/* set combined waveform which doesn't work on 6581 */ //udelay(50);
      for(i=0; i!=0xffff; ++i)
	if((sid_read(0x1b, s) & 0x80) != 0)
	  break;
      cw_unlock();

      sid_typ[s]=(i==0xffff)?SID_6581:SID_8580;
      printk(KERN_NOTICE DEVICE_NAME " : SID %i is %i\n", s, (i==0xffff)?6581:8580);
    }
  sid_reset();
}

int cwsid_write(struct file *file, const char *buf, size_t size, loff_t *pos)
{
  int max, i;

  if(file->f_pos >= sizeof(sid_buffer))
    return -ENOSPC;

  max=(file->f_pos+size>sizeof(sid_buffer))?(sizeof(sid_buffer)-file->f_pos):size;
  if(max<=0)
    return 0;
  if(max==1)
    {
      unsigned char v;
      if(get_user(v, buf) != 0)
	return -EFAULT;
      sid_buffer[file->f_pos]=v;
    }
  else
    {
      if(copy_from_user(sid_buffer + file->f_pos, buf, max) != 0)
	return -EFAULT;
    }

  cw_lock(CW_BANK_FLOPPY);
  for(i=file->f_pos; i<file->f_pos+max; i++)
    sid_write(i&0x1F, sid_buffer[i], i>>5);
  cw_unlock();

  file->f_pos += max;
  return max;
}

int cwsid_read(struct file *file, char *buf, size_t size, loff_t *pos)
{
  int max, i;

  max=(file->f_pos+size>sizeof(sid_buffer))?(sizeof(sid_buffer)-file->f_pos):size;

  cw_lock(CW_BANK_FLOPPY);
  for(i=file->f_pos; i<file->f_pos+max; i++)
    if((i&0x1F)>=0x19 && (i&0x1F)<=0x1C)
      sid_buffer[i]=sid_read(i&0x1F, i>>5);
  cw_unlock();

  if(copy_to_user(buf, sid_buffer+file->f_pos, max) != 0)
    return -EFAULT;

  file->f_pos += max;
  return max;
}

static int cwsid_opencount=0;
int cwsid_open(struct inode *inode, struct file *file)
{
  if(++cwsid_opencount == 1)
    {
      file->f_pos=0;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,8)
      file->f_mode |= FMODE_LSEEK | FMODE_PREAD | FMODE_PWRITE;
#endif
      sid_reset();
    }
  return 0;
}

int cwsid_close(struct inode *inode, struct file *file)
{
  if(--cwsid_opencount<=0)
    {
      sid_reset();
      cwsid_opencount=0;
    }
  return 0;
}

int cwsid_ioctl(struct inode *node, struct file *file, unsigned int cmd, unsigned long arg)
{
  switch(cmd)
    {
    case CWSID_IOCTL_RESET:
      sid_reset();
      return 0;

    case CWSID_IOCTL_SIDTYPE:
      {
	int i;
	get_user(i, (int*)arg);
	if(i>CWSID_MAX-1) i=CWSID_MAX-1;
	else if(i<0) i=0;
	put_user(sid_typ[i], (int*)arg);
	return 0;
      }

    case CWSID_IOCTL_CARDTYPE:
      put_user(CWSID_MAGIC_MK4, (int*)arg);
      return 0;

    case CWSID_IOCTL_PAL:
      cwsid_ntsc=0;
      return 0;

    case CWSID_IOCTL_NTSC:
      cwsid_ntsc=CWSID_NTSCBIT;
      return 0;

    case CWSID_IOCTL_DELAY:
      cwsid_delay=arg;
      return 0;
    }

  return -EINVAL;
}

loff_t cwsid_lseek(struct file *file, loff_t offset, int orig)
{
  const int maxsize=sizeof(sid_buffer);

  loff_t ret=-1;
  struct inode *inode = file->f_mapping->host;
  down(&inode->i_sem);

  switch(orig)
    {
#ifndef SEEK_SET
#define SEEK_SET 0
#endif
    case SEEK_SET:
      if(offset>=maxsize)
	{
	  file->f_pos=maxsize;
	  ret=-EINVAL;
	  break;
	}
      if(offset<0)
	{
	  file->f_pos=0;
	  ret=-EINVAL;
	  break;
	}
      ret=(file->f_pos = offset);
      break;

#ifndef SEEK_CUR
#define SEEK_CUR 1
#endif
    case SEEK_CUR:
      if(file->f_pos+offset >= maxsize)
	{
	  file->f_pos = maxsize;
	  ret=-EINVAL;
	  break;
	}
      if(file->f_pos+offset < 0)
	{
	  file->f_pos = 0;
	  ret=-EINVAL;
	  break;
	}
      ret=(file->f_pos += offset);
      break;

#ifndef SEEK_END
#define SEEK_END 2
#endif
    case SEEK_END:
      if(offset<0)
	{
	  file->f_pos = maxsize;
	  ret=-EINVAL;
	  break;
	}
      if(offset>=maxsize)
	{
	  file->f_pos = 0;
	  ret=-EINVAL;
	  break;
	}
      ret=(file->f_pos = maxsize-offset-1);
      break;

    default:
      ret=-EINVAL;
    }

  up(&inode->i_sem);
  return ret;
}

static struct file_operations cwsid_fops =
  {
    .owner = THIS_MODULE,
    .read = cwsid_read,
    .write = cwsid_write,
    .ioctl = cwsid_ioctl,
    .open = cwsid_open,
    .release = cwsid_close,
    .llseek = cwsid_lseek,
  };

static struct miscdevice cwsid_miscdevice =
  {
    .minor = CWSIDMINOR,
    .name = "sid",
    .fops = &cwsid_fops,
  };

static int initialized=0;

int __init cwsid_init(void)
{
  if(misc_register(&cwsid_miscdevice) < 0)
    {
      printk(KERN_ERR DEVICE_NAME ": register failed (already in use ?)\n");
      return -EIO;
    }

  sid_detect();

  printk(KERN_NOTICE DEVICE_NAME ": sid delay = %i\n", cwsid_delay);
  initialized=1;
  return 0;
}

void __exit cwsid_exit(void)
{
  if(initialized)
    {
      if(misc_deregister(&cwsid_miscdevice) < 0)
	printk(KERN_ERR DEVICE_NAME ": unregister failed\n");

      sid_reset();
      initialized=0;
    }
}
