/*
  sid (c) 1999-2004 by Dirk Jagdmann <doj@cubic.org>

  This file is part of sid.

  sid is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  sid is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with sid; if not, write to the Free Software Foundation, Inc.,
  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

  $Header: /code/cw/sid/c64.cc,v 1.21 2005/01/22 19:04:53 doj Exp $
*/

extern char *errstr;

#include <stdlib.h>

#include "c64.h"
#include "stilinfo.hpp"

const int C64::palfreq=985248;
const int C64::ntscfreq=1022700;

C64::C64() :
  cpu_(this)
{
  for(int i=0; i<SID_MAX_NUM; i++)
    sid[i].setNum(i);

  sidBase[0]=0xd400;
  sidBase[1]=-1000;
  sidBase[2]=-1000;
  sidBase[3]=-1000;
  sidNum_=1;

  reset();
}

const byte C64::getmem(const word addr)
{
  // sid
  for(int i=0; i<SID_MAX_NUM && i<sidNum_; i++)
    if(addr>=sidBase[i] && addr<sidBase[i]+0x20)
      return sid[i].get(addr&0x1f);

  if(watch(addr))
    cpu_.abort=true;

  return mem[addr];
}

void C64::setmem(const word addr, const byte value)
{
  if(addr>=0xD51D && addr<=0xD57F)
    {
      infostr.insert("poke to special SID sample register set 2");
      //printf("[%04X]=%02x\n", addr, value);
    }
  switch(addr)
    {
    case 0xD41D: infostr.insert("poke to sample Control"); break;
    case 0xD41E: case 0xD41F: infostr.insert("poke to sample Start Addr"); break;
    case 0xD43D: infostr.insert("poke to sample Length"); break;
    case 0xD43E: case 0xD43F: infostr.insert("poke to sample End Addr"); break;
    case 0xD45D: infostr.insert("poke to sample periodLo"); break;
    case 0xD45E: infostr.insert("poke to sample periodHi"); break;
    case 0xD45F: infostr.insert("poke to sample gap"); break;
    case 0xD47D: infostr.insert("poke to sample order"); break;
    case 0xD47E: case 0xD47F: infostr.insert("poke to sample repeat"); break;

    case 0x0314: case 0x0315: infostr.insert("poke to IRQ vector 2"); break;
    case 0x0316: case 0x0317: infostr.insert("poke to BRK vector 2"); break;
    case 0x0318: case 0x0319: infostr.insert("poke to NMI vector 2"); break;
    case 0xDC04: case 0xDC05: infostr.insert("poke to CIA1 Timer1"); CIAdelay=true; break;
    case 0xDC06: case 0xDC07: infostr.insert("poke to CIA1 Timer2"); break;
    case 0xDD04: case 0xDD05: infostr.insert("poke to CIA2 Timer1"); break;
    case 0xDD06: case 0xDD07: infostr.insert("poke to CIA2 Timer2"); break;
    case 0xFFFA: case 0xFFFB: infostr.insert("poke to NMI vector 1"); break;
    case 0xFFFC: case 0xFFFD: infostr.insert("poke to reset vector"); break;
    case 0xFFFE: case 0xFFFF: infostr.insert("poke to IRQ vector 1"); break;
    }

  for(int i=0;  i<sidNum_ && i<SID_MAX_NUM; ++i)
    if(addr>=sidBase[i] && addr<sidBase[i]+0x20)
      sid[i].set(addr&0x1F, value, cpu_.CPUcyclesTotal());

  if(watch(addr))
    cpu_.abort=true;

  mem[addr]=value;
}

void C64::reset()
{
  memset(mem, 0, 0xe000);
  memset(mem + 0xe000, 0x60, 0x2000); // Fill kernal ROM area with RTS
  mem[0]=0x2F;			// 6510 DDR
  mem[1]=0x37;			// 6510 PR
  mem[0x0314]=0x31; mem[0x0315]=0xea; // IRQ vector

  mem[0x0200]=0x60;		// RTS
  mem[0x0316]=0x00; mem[0x0316]=0x02; // BRK vector -> RTS
  mem[0x0317]=0x00; mem[0x0318]=0x02; // NMI vector -> RTS
  mem[0xFFFA]=0x00; mem[0xFFFB]=0x02; // NMI vector -> RTS
  mem[0xFFFC]=0x00; mem[0xFFFD]=0x02; // reset vector -> RTS
  mem[0xFFFE]=0x00; mem[0xFFFF]=0x02; // IRQ vector -> RTS

  cpu_.reset();
  resetSID();

  CIAdelay=false;
  pal_=true;
  setCIAtimer1();
}

void C64::resetSID()
{
  for(int i=0; i<SID_MAX_NUM; i++)
    sid[i].reset();
}

void C64::pal(const bool b)
{
  pal_=b;
  for(int i=0; i<sidNum_; i++)
    pal_?sid[i].pal():sid[i].ntsc();
}

void C64::mute()
{
  for(int i=0; i<sidNum_; i++)
    for(int j=0; j<3; j++)
      if((voicePauseStatus[i*3+j]=sid[i].voiceActive(j+1)))
	sid[i].toggle(j+1);
}

void C64::unmute()
{
  for(int i=0; i<sidNum_; i++)
    for(int j=0; j<3; j++)
      if(voicePauseStatus[i*3+j])
	sid[i].toggle(j+1);
}

bool C64::toggle(int v)
{
  switch(v)
    {
    case 1: return sid[0].toggle(1);
    case 2: return sid[0].toggle(2);
    case 3: return sid[0].toggle(3);
    case 4: return sid[1].toggle(1);
    case 5: return sid[1].toggle(2);
    case 6: return sid[1].toggle(3);
    case 7: return sid[2].toggle(1);
    case 8: return sid[2].toggle(2);
    case 9: return sid[2].toggle(3);
    case 0:
      for(v=0; v<sidNum_; v++)
	sid[v].toggleFilter();
      return true;
    }
  return false;
}

bool C64::voiceActive(const int v)
{
  switch(v)
    {
    case 1: return sid[0].voiceActive(1);
    case 2: return sid[0].voiceActive(2);
    case 3: return sid[0].voiceActive(3);
    case 4: return sid[1].voiceActive(1);
    case 5: return sid[1].voiceActive(2);
    case 6: return sid[1].voiceActive(3);
    case 7: return sid[2].voiceActive(1);
    case 8: return sid[2].voiceActive(2);
    case 9: return sid[2].voiceActive(3);
    case 0: return sid[0].voiceActive(0);
    }
  return false;
}

int C64::jsr(const word npc, const byte na, const byte nx, const byte ny, const int maxCycles)
{
  return cpu_.jsr(npc, na, nx, ny, maxCycles);
}

int C64::setSIDdevice(const char* device)
{
  for(int i=0; i<SID_MAX_NUM; i++)
    if(sid[i].open(device) < 0)
      return -1;
  return 0;
}

void C64::setVolume(const int v)
{
  for(int i=0; i<sidNum_; i++)
    sid[i].setVolume(v);
}

void C64::maxVolume(const int v)
{
  for(int i=0; i<sidNum_; i++)
    sid[i].maxVolume(v);
}

void C64::transpose(const double t)
{
  for(int i=0; i<sidNum_; i++)
    sid[i].transpose(t);
}

void C64::set(allSID& a)
{
  int i;
  for(i=0; i<sidNum_; i++)
    sid[i].set(a.sid[i]);
}

void C64::get(allSID& a)
{
  int i;
  for(i=0; i<sidNum_; i++)
    sid[i].get(a.sid[i]);
  for(; i<SID_MAX_NUM; i++)
    memset(&a.sid[i], 0, sizeof(SID));
}

void C64::setBase(const int s, const int b)
{
  if(s>=0 && s<SID_MAX_NUM)
    {
      sidBase[s]=b;
      if(b>=0)
	sidNum_=s+1;
    }
}

void C64::setGateBits(const bool b)
{
  for(int i=0; i<sidNum_; i++)
    sid[i].setGateBits(b);
}

bool C64::gateTriggered(const int v)
{
  if(v>=0 && v<sidNum_*3)
    return sid[v/3].gateTriggered((v%3)+1);
  return false;
}

int C64::delay()
{
  if(CIAdelay)
    {
      /*long long*/double a=1000000;
      a*=lohi(mem[0xDC04], mem[0xDC05])/*-0x100*/;
      a/=(pal_?palfreq:ntscfreq);
      /*a*=.8;*/
      return static_cast<int>(a);
    }

  return pal_?20000:16666;
}

void C64::setCIAtimer1()
{
  if(pal_)
    mem[0xDC04]=0x25, mem[0xDC05]=0x40;
  else
    mem[0xDC04]=0x95, mem[0xDC05]=0x42;
}

void C64::setCIAdelay(const bool b)
{
  CIAdelay=b;
  b?infostr.insert("enabling CIA delay"):infostr.insert("disabling CIA delay");
}

void C64::watchToggle(const word addr)
{
  if(watch(addr))
    {
      watchpoint.erase(addr);
      printf("clearing watchpoint at %04X\n", addr);
    }
  else
    {
      watchpoint.insert(addr);
      printf("setting watchpoint at %04X\n", addr);
    }
}

void C64::watchClear()
{
  watchpoint.clear();
}

void C64::watchPrint()
{
  printf("WatchPoints: ");
  for(watchpoint_t::iterator i=watchpoint.begin(); i!=watchpoint.end(); ++i)
    printf("%04X ", *i);
  printf("\n");
}

bool C64::watch(const word addr)
{
  return watchpoint.find(addr)!=watchpoint.end();
}

int C64::sidtype(int s) const
{
  return (s<0 || s>=sidNum_) ? SID_NONE : sid[s].type();
}

void C64::swapsid()
{
  MOS6581 s=sid[0];
  sid[0]=sid[1];
  sid[1]=s;
}
