/*
  sid (c) 1999-2004 by Dirk Jagdmann <doj@cubic.org>

  This file is part of sid.

  sid is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  sid is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with sid; if not, write to the Free Software Foundation, Inc.,
  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

  $Header: /code/cw/sid/main.cc,v 1.57 2005/02/25 09:51:27 doj Exp $
*/

#include <algorithm>
#include <fstream>
#include <iostream>
#include <iterator>
#include <map>
#include <string>

#include "average.hpp"
#include "c64.h"
#include "comment.h"
#include "configuration.hpp"
#include "load.hh"
#include "playlist.h"
#include "psid.h"
#include "sid.h"
#include "sidinfo.h"
#include "stil.h"
#include "stilinfo.hpp"
#include "time.hpp"
#include "trackinfo.hpp"

#include <curses.h>
#undef clear			// we don't need curses clear makro beyond this point
#include <errno.h>
#include <fcntl.h>
#include <sched.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <termios.h>
#include <unistd.h>

#define SHOWCYCLES 0

#ifdef USELIRC
#include <lirc/lirc_client.h>
static int lircfh=-1;
struct lirc_config *lircConfig=0;
#endif

#if defined(USETSID)
#include <tsid/tsid.h>
TSID *tsid=0;
#elif defined(USETSID2)
#include <tsid2/tsid2.h>
TSID2 *tsid2=0;
#endif

using namespace std;

enum event_t {
  eventNil=0,			// nothing
  eventExit,			// exit the program
  eventNextSong,		// play next song
  eventPrevSong,		// play previous song
  eventPauseToggle,		// toggle pause
  eventNextFile,		// load next file in argv
  eventPrevFile,		// load previous file in argv
  eventStopSong,		// stop the currently played song
  eventRandomFile,		// play a random file of the list
  eventToggle1,			// toggle Voice 1
  eventToggle2,			// toggle Voice 2
  eventToggle3,			// toggle Voice 3
  eventToggle4,			// toggle Voice 4
  eventToggle5,			// toggle Voice 5
  eventToggle6,			// toggle Voice 6
  eventToggle7,			// toggle Voice 7
  eventToggle8,			// toggle Voice 8
  eventToggle9,			// toggle Voice 9
  eventToggleF,			// toggle Filter
  eventPAL,			// switch to PAL speed
  eventNTSC,			// switch to NTSC speed
  eventDoubleSpeed,		// toggle normal, double, quad speed
  eventSkipBack,		// go some time into the past
  eventSkipForw,		// go some time into the future
  eventSkipBack2,
  eventSkipForw2,
  eventSkipBack3,
  eventSkipForw3,
  eventScrollspeed1,
  eventScrollspeed2,
  eventScrollspeed3,
  eventScrollspeed4,
  eventScrollspeed5,
  eventScrollspeed6,
  eventScrollspeed7,
  eventScrollspeed8,
  eventDebug,
  eventPlaylistUp,
  eventPlaylistDown,
  eventPlaylistSelect,
  eventTransposeUp,
  eventTransposeDown,
  eventPlayspeedUp,
  eventPlayspeedDown,
};

static WINDOW *mainw=0, *timew=0;
#if SHOWCYCLES
static WINDOW *cyclew=0;
#endif
static int timewWidth;
static volatile event_t EVENT=eventNil;
static psid *sidinfo=0;
static C64 *c64=0;
char *errstr=0;
static STIL *stil=0;
static volatile bool cursessetup=true, cursesredraw=true;
static int doublespeed=1;
static Playlist playlist;
static double transpose=1;
static double playspeed=1;

static void fixPlayAddr(psid *ps, C64 *comp)
{
  // Replay routine address is given by interrupt vector
  if(ps->playAddr == 0)
    {
      // Kernal ROM switched in
      if(comp->getmem(1) & 2)
	ps->playAddr = lohi(comp->getmem(0x0314), comp->getmem(0x0315));
      // Kernal ROM switched out
      else
	ps->playAddr = lohi(comp->getmem(0xfffe), comp->getmem(0xffff));
    }
}

static void showversion()
{
  cout << "INFO: sid " VERSION " (c) 1999-2005 by Dirk Jagdmann and Tammo Hinrichs" << endl;
}

static void showhelp()
{
  showversion();
  cout << "CMDL: sid [options] filenames" << endl;
  cout << "      options: -a          advance to next song upon end of song" << endl;
  cout << "               -b directoy set HVSC base directory" << endl;
  cout << "               -c num      maximum number of CPU cycles" << endl;
  cout << "               -d dev      use dev as output device" << endl;
  cout << "               -h          print help" << endl;
  cout << "               -r          randomize playlist" << endl;
  cout << "               -s num      play song num" << endl;
  cout << "               -t seconds  set timeout to seconds" << endl;
  cout << "               -v          print version" << endl;
  cout << "               -w          swap SID0 and SID1" << endl;
  cout << "               -W          play on SID0 and SID1" << endl;
  cout << "               -2          set address of second SID" << endl;
  cout << "               -3          set address of third SID" << endl;
  cout << "               -4          set address of fourth SID" << endl;
  cout << "               -D          start with debugger" << endl;
  cout << "KEYS: n - play next song" << endl;
  cout << "      b - play previous song" << endl;
  cout << "      + - play next file from playlist" << endl;
  cout << "      - - play previous file from playlist" << endl;
  cout << "      r - play random file from playlist" << endl;
  cout << "space,p - pause" << endl;
  cout << "1,2,3,4,5,6,7,8,9 - toggle Voice" << endl;
  cout << "      f - toggle Filter" << endl;
  cout << "      N - NTSC speed" << endl;
  cout << "      P - PAL speed" << endl;
  cout << "      d - toggle between normal,double,quad speed" << endl;
  cout << "  esc,q - quit program" << endl;
  cout << " a,home - restart song" << endl;
  cout << "   left - skip back 1s" << endl;
  cout << "  right - skip forward 1s" << endl;
  cout << "   pgup - skip back 10s" << endl;
  cout << " pgdown - skip forward 10s" << endl;
  cout << "  F1-F8 - change scroll speed of track display" << endl;
  cout << "      D - start debugger" << endl;
  cout << "     up - move one up in playlist" << endl;
  cout << "   down - move one down in playlist" << endl;
  cout << "enter,KP5,s - select current song from playlist" << endl;
  cout << " F9,y,z - transpose one halftone down" << endl;
  cout << "  F10,x - transpose one halftone up" << endl;
  cout << "  F11,c - playspeed down" << endl;
  cout << "  F12,v - playspeed up" << endl;
}

static void curses_init()
{
  WINDOW *w=initscr();
  if(!w)
    {
      errstr="could not init curses";
      exit(1);
    }
  savetty();
  //start_color();
  noecho();
  raw();
  curs_set(0);
  keypad(w, TRUE);
}

static void curses_exit()
{
  resetty();
  endwin();
}

static void cleanup()
{
  const int e=errno;

  curses_exit();

  delete c64, c64=0;
  delete stil, stil=0;

#ifdef USELIRC
  if(lircfh >= 0)
    {
      if(lircConfig)
	lirc_freeconfig(lircConfig);
      lirc_deinit();
    }
#endif

#if defined(USETSID)
  delete tsid, tsid=0;
#elif defined(USETSID2)
  delete tsid2, tsid2=0;
#endif

  if((errno=e))
    perror("misc error");

  if(errstr)
    cout << endl << errstr << endl;
  else if(sidinfo)
    {
      commentPrint(sidinfo);
      delete sidinfo, sidinfo=0;
    }

}

static int windows_init(C64 *c64)
{
  int xoffset=0, yoffset=0;

  // setup track window
  if(trackinfo_init(xoffset, yoffset, c64) < 0)
    return -1;

  // setup time window
  if(timew)
    delwin(timew);

  timewWidth=COLS-xoffset;
  timew=newwin(1, timewWidth, 0, xoffset);
  if(!timew)
    {
      errstr="could not init timew. maybe terminal width too small?";
      return -1;
    }
  yoffset=1;

  // setup main window
  if(mainw)
    delwin(mainw);

  //const int mainwTopLeftX=xoffset;
  const int mainwTopLeftY=yoffset;
  mainw=newwin(6, 45, yoffset, xoffset);
  if(!mainw)
    {
      errstr="could not init mainw. maybe terminal width too small?";
      return -1;
    }
  keypad(mainw, TRUE);
  yoffset+=6;
  const int mainwBotRightX=xoffset+45;
  //const int mainwBotRightY=yoffset;

#if SHOWCYCLES
  // setup cycle window
  if(cyclew)
    delwin(cyclew);
  cyclew=newwin(1, 45, yoffset, xoffset);
  if(!cyclew)
    {
      errstr="could not init cyclew. maybe terminal with too small?";
      return -1;
    }
  yoffset++;
  mvwprintw(cyclew, 0, 0, "cycles:");
#endif

  // setup sid window
  const int sidwTopLeftX=xoffset;
  //const int sidwTopLeftY=yoffset;
  sidinfo_init(xoffset, yoffset, c64);
  //const int sidwBotRightX=xoffset;
  const int sidwBotRightY=yoffset;

  // setup playlist window
  xoffset=mainwBotRightX+1;
  yoffset=mainwTopLeftY;
  playlist.init(xoffset, yoffset, sidwBotRightY-yoffset);
  playlist.playinfo_print();

  // setup stil window
  xoffset=sidwTopLeftX;
  yoffset=sidwBotRightY;
  if(stilinfo_init(xoffset, yoffset) < 0)
    return -1;

  return 0;
}

#if SHOWCYCLES
static void cycleinfo_print(int c)
{
  wattron(mainw, A_BOLD);
  mvwprintw(cyclew, 0, 8, "%i       ", c);
  wattroff(mainw, A_BOLD);
}
#endif

static void maininfo_print(psid *sidinfo, const std::string& filename)
{
  // should transpose be displayed?
  const bool tr=(transpose>1.01 || transpose<0.99)?true:false;

  // should speed be displayed?
  const bool sp=(playspeed>1.01 || playspeed<0.99)?true:false;

  // width for tunenumber
  int wd=4;
  if(sidinfo->songTotal>=10)
    wd+=2;
  if(sidinfo->songTotal>=100)
    wd+=2;

  // truncate filename if necessary
  std::string displayfilename=filename;
  const int max=timewWidth-8-wd-(tr?6:0)-(sp?6:0); // the 8 is for the initial time
  if(max>0 && static_cast<int>(displayfilename.size())>max)
    displayfilename.erase(0, displayfilename.size()-max);

  // print into timew
  werase(timew);
  mvwprintw(timew, 0, 8, "%s %i/%i",
	    displayfilename.c_str(),
	    sidinfo->songPlay+1,
	    sidinfo->songTotal
	    );

  if(tr)
    wprintw(timew, " T%03i%%", static_cast<int>(transpose*100.0));

  if(sp)
    wprintw(timew, " S%03i%%", static_cast<int>(playspeed*100.0));

  // print into mainw
  int y=0;
  werase(mainw);

  mvwprintw(mainw, y++, 0, "name:      ");
  wattron(mainw, A_BOLD);
  waddstr(mainw, sidinfo->name);
  wattroff(mainw, A_BOLD);

  mvwprintw(mainw, y++, 0, "author:    ");
  wattron(mainw, A_BOLD);
  waddstr(mainw, sidinfo->author);
  wattroff(mainw, A_BOLD);

  mvwprintw(mainw, y++, 0, "copyright: ");
  wattron(mainw, A_BOLD);
  waddstr(mainw, sidinfo->copyright);
  wattroff(mainw, A_BOLD);

  mvwprintw(mainw, y++, 0, "flags:     ");
  wattron(mainw, A_BOLD);
  if(sidinfo->MUSplayer) waddstr(mainw, "MUS player ");
  if(sidinfo->PlaySIDSamples) waddstr(mainw, "PlaySID samples ");
  if(sidinfo->pal)
    waddstr(mainw, "PAL ");
  else
    waddstr(mainw, "NTSC ");
  if(sidinfo->mos6581) waddstr(mainw, "MOS6581 ");
  if(sidinfo->mos8580) waddstr(mainw, "MOS8580 ");
  wattroff(mainw, A_BOLD);

  wattron(mainw, A_UNDERLINE);
  mvwprintw(mainw, y++, 0, "v off loadA initA playA speed flags reserved");
  wattroff(mainw, A_UNDERLINE);

  wattron(mainw, A_BOLD);
  mvwprintw(mainw, y++, 0, "%i $%02x $%04x $%04x $%04x %05i $%04x %08x",
	    (int)sidinfo->version,
	    (int)sidinfo->dataOffset,
	    (int)sidinfo->loadAddr,
	    (int)sidinfo->initAddr,
	    (int)sidinfo->playAddr,
	    (int)sidinfo->speed,
	    (int)sidinfo->flags,
	    (int)sidinfo->reserved
	    );
  wattroff(mainw, A_BOLD);

  wnoutrefresh(mainw);
}

static void USR1(int)
{
  EVENT=eventNextSong;
}

static void USR2(int)
{
  EVENT=eventPrevSong;
}

static void TERM(int)
{
  EVENT=eventExit;
}

/* signal handler for SIGWINCH */
static void RESIZE(int sig)
{
  struct winsize size;
  if(ioctl(0, TIOCGWINSZ, &size) < 0)
    return;

  resizeterm(size.ws_row, size.ws_col);
  cursessetup=true;
}

static std::string md5str(const std::string& fn)
{
  char buf[33];
  FILE *p=popen((std::string("md5sum ")+fn).c_str(), "r");
  if(!p)
    return "";
  fgets(buf, 32, p);
  pclose(p);
  buf[32]=0;
  return buf;
}

int main(const int argc, char * const *argv)
{
  unsigned long timeout=~0;
  char *device=0;
  int song=-1, sid2=0, sid3=0, sid4=0, fhmax=0, maxCycles=1000000/50;
  bool randomize=false, startupdebugger=false, brutalpolling=false, swapsid=false;
  doj::Average<int, 50> avgCycles;
  event_t endOfSongEvent=eventStopSong;

  atexit(cleanup);
  srand(time(NULL));

  // check command line
  if(argc<2)
    {
      showhelp();
      exit(1);
    }

  STILbase=getenv("HVSC_DIR");

  doj::Configuration conf;
  conf.load("/etc/sid.conf");
  conf.updateString("device", device);
  conf.updateInt("timeout", timeout);
  conf.updateBool("randomize", randomize);
  conf.updateString("stilbase", STILbase);

  int c=0;
  while(c!=EOF)
    {
      c=getopt(argc, argv, "2:3:4:ab:c:d:hi:prs:t:vwWD");
      switch(c)
	{
	case 'a': endOfSongEvent=eventNextSong; break;
	case 'b': STILbase=optarg; break;
	case 'c': maxCycles=strtol(optarg, NULL, 0); break;
	case 'd': device=optarg; break;
	case 'h': showhelp(); exit(0);
	case 'p': brutalpolling=true; break;
	case 'r': randomize=true; break;
	case 's': song=strtol(optarg, NULL, 0); if(song>0) song--; break;
	case 't': timeout=strtol(optarg, NULL, 0); break;
	case 'v': showversion(); exit(0);
	case 'w': swapsid=true; break;
	case 'W': sid2=0xD400; break;
	case '2': sid2=strtol(optarg, NULL, 0); break;
	case '3': sid3=strtol(optarg, NULL, 0); break;
	case '4': sid4=strtol(optarg, NULL, 0); break;
	case 'D': startupdebugger=true; break;
	}
    }

  if(STILbase)
    {
      string base=absFilename(STILbase);

      stil=new STIL;
      if(!stil->setBaseDir(base.c_str()))
	{
	  cerr << "could not set STIL basedir: " << stil->getErrorStr() << endl;
	  return 1;
	}

#if defined(USETSID)
      tsid=new TSID;
      if(!tsid->testHVSC(base.c_str()))
	{
	  cerr << base << " is not a valid HVSC dir" << endl;
	  return 1;
	}
      if(!tsid->setBaseDir(base.c_str()))
	{
	  cerr << "could not set TSID HVSC basedir: " << tsid->getError() << endl;
	  return 1;
	}
#elif defined(USETSID2)
      tsid2=new TSID2;
#endif

    }

  // initialize playlist
  for(int i=optind; i<argc; i++)
    playlist.add(argv[i]);

  if(randomize)
    playlist.randomize();

  // install signal handlers
  signal(SIGUSR1, USR1);
  signal(SIGUSR2, USR2);
  signal(SIGTERM, TERM);
  signal(SIGINT, TERM);
  signal(SIGWINCH, RESIZE);

#ifdef USELIRC
  map<string, event_t> lircMap;

  // set up device
  lircfh=lirc_init("sid", 1);
  if(lircfh < 0)
    cerr << "could not init LIRC" << endl;
  else
    {
      if(lircfh>fhmax)
	fhmax=lircfh;

      // read config
      if(lirc_readconfig(NULL /*configfile*/, &lircConfig, NULL) < 0)
	{
	  // if there is no config file, we don't want to use lirc
	  close(lircfh);
	  lircfh=-1;
	}
      else
	{
	  // set up config table
	  lircMap["next"]=eventNextSong;
	  lircMap["prev"]=eventPrevSong;
	  lircMap["pause"]=eventPauseToggle;
	  lircMap["quit"]=eventExit;
	  lircMap["nextfile"]=eventNextFile;
	  lircMap["prevfile"]=eventNextFile;
	  lircMap["randfile"]=eventRandomFile;
	  lircMap["1"]=eventToggle1;
	  lircMap["2"]=eventToggle2;
	  lircMap["3"]=eventToggle3;
	  lircMap["4"]=eventToggle4;
	  lircMap["5"]=eventToggle5;
	  lircMap["6"]=eventToggle6;
	  lircMap["7"]=eventToggle7;
	  lircMap["8"]=eventToggle8;
	  lircMap["9"]=eventToggle9;
	  lircMap["f"]=eventToggleF;
	}
    }
#endif

  // set up key2event table
  map<int, event_t> keyMap;
  keyMap[' ']=eventPauseToggle;
  keyMap['+']=eventNextFile;
  keyMap['-']=eventPrevFile;
  keyMap['1']=eventToggle1;
  keyMap['2']=eventToggle2;
  keyMap['3']=eventToggle3;
  keyMap['4']=eventToggle4;
  keyMap['5']=eventToggle5;
  keyMap['6']=eventToggle6;
  keyMap['7']=eventToggle7;
  keyMap['8']=eventToggle8;
  keyMap['9']=eventToggle9;
  keyMap['a']=eventStopSong;
  keyMap['b']=eventPrevSong;
  keyMap['c']=eventPlayspeedDown;
  keyMap['d']=eventDoubleSpeed;
  keyMap['f']=eventToggleF;
  keyMap['n']=eventNextSong;
  keyMap['p']=eventPauseToggle;
  keyMap['q']=eventExit;
  keyMap['r']=eventRandomFile;
  keyMap['s']=eventPlaylistSelect;
  keyMap['v']=eventPlayspeedUp;
  keyMap['x']=eventTransposeUp;
  keyMap['y']=eventTransposeDown;
  keyMap['z']=eventTransposeDown;
  keyMap['D']=eventDebug;
  keyMap['N']=eventNTSC;
  keyMap['P']=eventPAL;
  keyMap[KEY_HOME]=eventStopSong;
  keyMap[KEY_A1]=eventStopSong;
  keyMap[KEY_LEFT]=eventSkipBack;
  keyMap[KEY_RIGHT]=eventSkipForw;
  keyMap[KEY_NPAGE]=eventSkipBack2;
  keyMap[KEY_C3]=eventSkipBack2;
  keyMap[KEY_PPAGE]=eventSkipForw2;
  keyMap[KEY_A3]=eventSkipForw2;
  keyMap[KEY_F(1)]=eventScrollspeed1;
  keyMap[KEY_F(2)]=eventScrollspeed2;
  keyMap[KEY_F(3)]=eventScrollspeed3;
  keyMap[KEY_F(4)]=eventScrollspeed4;
  keyMap[KEY_F(5)]=eventScrollspeed5;
  keyMap[KEY_F(6)]=eventScrollspeed6;
  keyMap[KEY_F(7)]=eventScrollspeed7;
  keyMap[KEY_F(8)]=eventScrollspeed8;
  keyMap[KEY_F(9)]=eventTransposeDown;
  keyMap[KEY_F(10)]=eventTransposeUp;
  keyMap[KEY_F(11)]=eventPlayspeedDown;
  keyMap[KEY_F(12)]=eventPlayspeedUp;
  keyMap[KEY_DOWN]=eventPlaylistDown;
  keyMap[KEY_UP]=eventPlaylistUp;
  keyMap[KEY_B2]=eventPlaylistSelect;
  keyMap[KEY_ENTER]=eventPlaylistSelect;

  // initialize the C64
  c64=new C64;
  if(c64->setSIDdevice(device) < 0)
    {
      errstr="main(): could not open SID device";
      exit(1);
    }
  // enable additional SIDs
  if(sid2)
    {
      c64->setBase(1, sid2);
      if(sid3)
	{
	  c64->setBase(2, sid3);
	  if(sid4)
	    c64->setBase(3, sid4);
	}
    }
  if(swapsid)
    c64->swapsid();
  for(int i=0; i!=c64->sidNum(); ++i)
    {
      std::clog << "SID " << i << " : ";
      switch(c64->sidtype(i))
	{
	default:
	case SID_NONE: std::clog << " none\n" << std::flush; break;
	case SID_6581: std::clog << " 6581\n" << std::flush; break;
	case SID_8580: std::clog << " 8580\n" << std::flush; break;
	}
    }
  if(c64->sidtype(0) == SID_NONE)
    {
      errstr="SID0 not present";
      exit(1);
    }

  // become a realtime process
  uid_t uid=getuid();
  if(uid!=0)
    {
      uid=geteuid();
      if(uid==0)
	{
	  setuid(0);
	  //seteuid(0);
	}
    }

  uid=getuid();
  if(uid==0)
    {
      struct sched_param sp; sp.sched_priority=sched_get_priority_max(SCHED_FIFO);
      if(sched_setscheduler(0, SCHED_FIFO, &sp) < 0)
	{
	  errstr="main(): could not set realtime scheduling";
	  exit(1);
	}
    }
  else
    errstr="could not set realtime scheduling. Please run me as root or setuid root this file";

  // initialize curses
  curses_init();

  while(EVENT!=eventExit)
    {
      string filename=playlist.current();
      if(song==-1 && playlist.currentsong()>=0)
	song=playlist.currentsong();

      // check if it is really a file
      struct stat statbuf;
      if(stat(filename.c_str(), &statbuf) < 0)
	{
	  if(playlist.size()>1)
	    playlist.next();
	  else
	    EVENT=eventExit;
	  continue;
	}

      // load sid file
      sidinfo=load(filename, c64);
      if(sidinfo==0)
	{
	  if(playlist.size()>1)
	    playlist.next();
	  else
	    EVENT=eventExit;
	  continue;
	}

      // setup siddump buffer
      std::vector<allSID> siddump;
      siddump.reserve(50*60*10); // reserve space for ~10 minutes
      int playpos=0;

      // set start song
      if(song==-2)
	{
	  // set the last song in the file
	  if(sidinfo->songTotal-1 > 0)
	    song=sidinfo->songPlay=sidinfo->songTotal-1;
	  else
	    song=sidinfo->songPlay=0;
	}
      else if(song>=0 && song<sidinfo->songTotal)
	sidinfo->songPlay=song;
      else
	{
	  if(sidinfo->songStart-1 > 0)
	    song=sidinfo->songPlay=sidinfo->songStart-1;
	  else
	    song=sidinfo->songPlay=0;
	}

      // initialize C64 and other things
      infostr.clear();
      c64->pal(sidinfo->pal);
      c64->setCIAtimer1();
      if(sidinfo->speed&(1<<sidinfo->songPlay))
	c64->setCIAdelay(true);
      c64->resetSID();
      if(startupdebugger)
	EVENT=eventDebug;
      else
	{
	  cursessetup=true;
	  EVENT=eventNil;
	  c64->setVolume(15); // set volume to max, because some sidtunes do not set volume
	  c64->jsr(sidinfo->initAddr, sidinfo->songPlay, 0, 0, maxCycles*10);
	  fixPlayAddr(sidinfo, c64);
	  playpos=0;
	  siddump.clear();
	}

      // the event loop
      unsigned long starttime=timeGetTime(); // in ms
      unsigned long currenttime=0; // in s
      unsigned long c64timeStart=timeGetTime();	// in ms
      unsigned long c64timeStop=c64timeStart; // in ms
      unsigned long fadeoutstarttime=0;	// in ms
      int timeadd=0;		// time correction in s
      bool pause=false;
      while(EVENT!=eventExit && EVENT!=eventStopSong)
	{
	  // check if curses needs to be setuped
	  if(cursessetup)
	    {
	      if(windows_init(c64) == 0)
		{
		  cursessetup=false;
		  cursesredraw=true;
		}
	      else
		{
		  EVENT=eventExit;
		  continue;
		}
	    }
	  // check if new information needs to be displayed
	  if(cursesredraw)
	    {
	      cursesredraw=false;

	      maininfo_print(sidinfo, filename);
	      stilinfo_print(stil, filename, sidinfo->songPlay+1);
	    }

	  // update clock
	  if(!pause)
	    currenttime=(timeGetTime()-starttime)/1000 + timeadd;

	  // check timeout
	  if(currenttime>=timeout)
	    EVENT=eventNextSong;

	  const int fadeouttime=3;
	  if(currenttime>=timeout-fadeouttime)
	    {
	      if(fadeoutstarttime==0)
		fadeoutstarttime=timeGetTime();
	      c64->maxVolume(15-static_cast<int>(static_cast<double>(timeGetTime()-fadeoutstarttime)/1000.0/static_cast<double>(fadeouttime)*15.0));
	    }

	  switch(EVENT)
	    {
	    case eventNextSong:
	      if(sidinfo->songPlay < sidinfo->songTotal - 1)
		{
		  song++;
		  EVENT=eventStopSong;
		}
	      else
		{
		  song=0;
		  goto eventNextFile_l;
		}
	      break;

	    case eventPrevSong:
	      if(sidinfo->songPlay > 0)
		{
		  song--;
		  EVENT=eventStopSong;
		}
	      else
		{
		  song=-2;
		  goto eventPrevFile_l;
		}
	      break;

	    case eventPauseToggle:
	      pause=!pause;
	      if(pause)
		{
		  c64->mute();
		  timeadd=currenttime;
		}
	      else
		{
		  c64->unmute();
		  starttime=timeGetTime();
		}
	      EVENT=eventNil;
	      break;

	    case eventNextFile:
 	      song=-1;
	    eventNextFile_l:
	      playlist.next();
	      EVENT=eventStopSong;
	      break;

	    case eventPrevFile:
 	      song=-1;
	    eventPrevFile_l:
	      playlist.previous();
	      EVENT=eventStopSong;
	      break;

	    case eventRandomFile:
	      playlist.random();
	      song=-1;
	      EVENT=eventStopSong;
	      break;

	    case eventToggle1:
	      c64->toggle(1);
	      EVENT=eventNil;
	      break;
	    case eventToggle2:
	      c64->toggle(2);
	      EVENT=eventNil;
	      break;
	    case eventToggle3:
	      c64->toggle(3);
	      EVENT=eventNil;
	      break;
	    case eventToggle4:
	      c64->toggle(4);
	      EVENT=eventNil;
	      break;
	    case eventToggle5:
	      c64->toggle(5);
	      EVENT=eventNil;
	      break;
	    case eventToggle6:
	      c64->toggle(6);
	      EVENT=eventNil;
	      break;
	    case eventToggle7:
	      c64->toggle(7);
	      EVENT=eventNil;
	      break;
	    case eventToggle8:
	      c64->toggle(8);
	      EVENT=eventNil;
	      break;
	    case eventToggle9:
	      c64->toggle(9);
	      EVENT=eventNil;
	      break;
	    case eventToggleF:
	      c64->toggle(0);
	      EVENT=eventNil;
	      break;

	    case eventNTSC:
	      c64->pal(false);
	      //sidinfo->ntsc=true;
	      //sidinfo->pal=false;
	      //cursesredraw=true;
	      EVENT=eventNil;
	      break;
	    case eventPAL:
	      c64->pal(true);
	      //sidinfo->ntsc=false;
	      //sidinfo->pal=true;
	      //cursesredraw=true;
	      EVENT=eventNil;
	      break;

	    case eventSkipBack:
	      c=1;
	    eventSkipBack_l:
	      playpos-=((sidinfo->pal)?50:60)*c;
	      if(playpos<0)
		EVENT=eventStopSong;
	      else
		{
		  timeadd-=c;
		  EVENT=eventNil;
		}
	      break;

	    case eventSkipForw:
	      c=1;
	    eventSkipForw_l:
	      playpos+=((sidinfo->pal)?50:60)*c;
	      timeadd+=c;
	      EVENT=eventNil;
	      break;

	    case eventSkipBack2:
	      c=10;
	      goto eventSkipBack_l;

	    case eventSkipForw2:
	      c=10;
	      goto eventSkipForw_l;

	    case eventSkipBack3:
	      c=60;
	      goto eventSkipBack_l;

	    case eventSkipForw3:
	      c=60;
	      goto eventSkipForw_l;

	    case eventDoubleSpeed:
	      doublespeed*=2;
	      if(doublespeed>4)
		doublespeed=1;
	      EVENT=eventNil;
	      break;

	    case eventDebug:
	      curses_exit();
	      if(c64->debug(sidinfo->playAddr, sidinfo->songPlay, 0, 0) < 0)
		{
		  EVENT=eventExit;
		  continue;
		}
	      curses_init();
	      cursessetup=true;
	      EVENT=eventNil;
	      break;

	    case eventPlaylistUp:
	      playlist.up();
	      EVENT=eventNil;
	      break;
	    case eventPlaylistDown:
	      playlist.down();
	      EVENT=eventNil;
	      break;
	    case eventPlaylistSelect:
	      playlist.select();
	      song=-1;
	      EVENT=eventStopSong;
	      break;

	    case eventScrollspeed1: trackinfo_scrollspeed(1); EVENT=eventNil; break;
	    case eventScrollspeed2: trackinfo_scrollspeed(2); EVENT=eventNil; break;
	    case eventScrollspeed3: trackinfo_scrollspeed(3); EVENT=eventNil; break;
	    case eventScrollspeed4: trackinfo_scrollspeed(4); EVENT=eventNil; break;
	    case eventScrollspeed5: trackinfo_scrollspeed(5); EVENT=eventNil; break;
	    case eventScrollspeed6: trackinfo_scrollspeed(6); EVENT=eventNil; break;
	    case eventScrollspeed7: trackinfo_scrollspeed(7); EVENT=eventNil; break;
	    case eventScrollspeed8: trackinfo_scrollspeed(8); EVENT=eventNil; break;

	    case eventTransposeUp:
	      transpose*=WURZ12OF2;
	      c64->transpose(transpose);
	      cursesredraw=true;
	      EVENT=eventNil;
	      break;
	    case eventTransposeDown:
	      transpose/=WURZ12OF2;
	      c64->transpose(transpose);
	      cursesredraw=true;
	      EVENT=eventNil;
	      break;

	    case eventPlayspeedUp:
	      playspeed+=.1;
	      cursesredraw=true;
	      EVENT=eventNil;
	      break;
	    case eventPlayspeedDown:
	      playspeed-=.1;
	      cursesredraw=true;
	      EVENT=eventNil;
	      break;

	    case eventNil:
	    default:
	      if(!pause)
		{
		  allSID allsid;
		  int size=siddump.size();
		  int c=0;
		  if(playpos==size)
		    {
		      c=c64->jsr(sidinfo->playAddr, sidinfo->songPlay, 0, 0, maxCycles);
		      c64->get(allsid);
		      siddump.push_back(allsid);
		      playpos++;
		    }
		  else if(playpos<size)
		    {
		      allsid=siddump[playpos++];
		      c64->set(allsid);
		    }
		  else
		    {
		      while(size<playpos)
			{
			  c=c64->jsr(sidinfo->playAddr, sidinfo->songPlay, 0, 0, maxCycles);
			  c64->get(allsid);
			  siddump.push_back(allsid);
			  size++;
			}
		    }
		  avgCycles.push(c);
#if SHOWCYCLES
		  cycleinfo_print(c);
		  wnoutrefresh(cyclew);
#endif
		}
	    }

	  // update curses
	  mvwprintw(timew, 0, 0, "%4i:%02i", (int)(currenttime/60), (int)(currenttime%60));
	  wnoutrefresh(timew);

	  sidinfo_print(c64);
	  if(trackinfo_print(c64) < 0)
	    {
	      EVENT=eventExit;
	      continue;
	    }
	  stilinfo_printinfostr();

	  doupdate();

	  // initialize the read set
	  fd_set readset;
	  FD_ZERO(&readset);
	  FD_SET(0, &readset);		// stdin

#ifdef USELIRC
	  if(lircfh >= 0)
	    FD_SET(lircfh, &readset);
#endif

	  // stop time needed for calculating current frame
	  c64timeStop=timeGetTime();

	  // set timeout
	  int delay=static_cast<int>(static_cast<double>(c64->delay() / doublespeed) / playspeed);
	  delay-=(c64timeStop-c64timeStart)*1000;
	  if(delay<0)
	    delay=0;

	  if(brutalpolling)
	    {
	      delay/=1000;	// convert delay to ms
	      // poll delay loop, yeah ugly
	      const int start=timeGetTime();
	      while(static_cast<int>(timeGetTime())-start < delay) ;
	      // select() call should not block
	      delay=0;
	    }

	  struct timeval tv;
	  tv.tv_sec = 0;
	  tv.tv_usec = delay;

	  // select will exit when
	  // 1) key was pressed on stdin
	  // 2) lirc reports an event
	  // 3) after timeout (when no interrupts available)
	  select(fhmax+1, &readset, NULL, NULL, pause?NULL:&tv);

	  // the new frame starts processing here
	  c64timeStart=timeGetTime();

	  // get a key from keyboard
	  if(FD_ISSET(0, &readset))
	    EVENT=keyMap[wgetch(mainw)];

#ifdef USELIRC
	  if(lircfh >= 0 && FD_ISSET(lircfh, &readset))
	    {
	      char *c=0;
	      if(lirc_nextcode(&c) < 0)
		{
		  errstr="main(): error in lirc_nextcode()";
		  exit(1);
		}

	      if(c)
		{
		  char *s=0;
		  if(lirc_code2char(lircConfig, c, &s) < 0)
		    {
		      errstr="main(): error in lirc_code2char()";
		      exit(1);
		    }

		  if(s)
		    EVENT=lircMap[s];

		  free(c);
		}

	    }
#endif

	  // check for end of song
	  if(avgCycles.full())
	    {
	      if(avgCycles.avg()<=100)
		EVENT=endOfSongEvent;
	      avgCycles.erase();
	    }

	} // while(EVENT!=eventExit && EVENT!=eventStopSong)

      if(sidinfo)
	{
#if defined(USETSID)
	  if(currenttime>=MIN_LISTEN_TIME)
	    tsid->addTime(currenttime, sidinfo->songPlay+1, absFilename(filename).c_str());
#elif defined(USETSID2)
	  if(currenttime>=MIN_LISTEN_TIME)
	    {
	      std::string f=absFilename(filename);
	      tsid2->addTime(md5str(f).c_str(), f.c_str(), currenttime, sidinfo->songPlay + 1);
	    }
#endif
	  delete sidinfo, sidinfo=0;
	}

    } // while(EVENT!=eventExit)

  return 0;
}
